"""Enumerated data types used by the Halo Infinite API."""

import logging
from enum import Enum, IntEnum, StrEnum

logger = logging.getLogger(__name__)


class AssetHome(IntEnum):
    """Source of an asset."""

    UNKNOWN = 0
    """Unknown asset source."""
    STUDIO = 1
    """Developer-sourced asset."""
    PLAYER = 2
    """Player-sourced asset."""


class AssetKind(IntEnum):
    """Types of assets used by Halo Infinite."""

    UNKNOWN = 0
    """Unknown asset type."""
    FILM = 1
    """A film asset."""
    MAP = 2
    """A map asset."""
    PLAYLIST = 3
    """A matchmaking playlist asset."""
    PREFAB = 4
    """A prefabricated object asset."""
    TEST_ASSET = 5
    """A test asset."""
    UGC_GAME_VARIANT = 6
    """A user-generated content game mode asset."""
    MAP_MODE_PAIR = 7
    """A map and game mode pair asset."""
    PROJECT = 8
    """A project asset."""
    MANIFEST = 9
    """A manifest asset."""
    ENGINE_GAME_VARIANT = 10
    """An engine game mode asset."""


class BotDifficulty(IntEnum):
    """Bot difficulty levels as encoded in stats responses."""

    UNKNOWN = -1
    """Unknown bot difficulty."""
    MARINE = 1
    """Marine bots. 2nd lowest difficulty."""
    ODST = 2
    """ODST bots. 2nd highest difficulty."""
    SPARTAN = 3
    """Spartan bots. Highest difficulty."""
    RECRUIT = 4
    """Recruit bots. Lowest difficulty."""
    ADAPTIVE = 5
    """Adaptive bots. Difficulty changes based on player performance."""


class CloneBehavior(IntEnum):
    """Permission levels for cloning assets."""

    DEFAULT = 0
    """Default clone behavior."""
    TEMPLATE = 1
    """Clone as a template."""
    PROHIBITED = 2
    """Cloning is prohibited."""


class FilmChunkType(IntEnum):
    """Types of saved film chunks."""

    NONE = 0
    """No film chunk type."""
    FILM_HEADER = 1
    """Film header chunk."""
    REPLICATION_DATA = 2
    """Replication data chunk."""
    HIGHLIGHT_EVENTS = 3
    """Highlight events chunk."""


class FilmStatus(IntEnum):
    """Status of a saved film."""

    UNKNOWN = 0
    """Unknown film status."""
    COMPLETE = 1
    """Film is complete."""
    ONGOING = 2
    """Film is ongoing."""
    RECOVERED = 3
    """Film was recovered."""


class GameVariantCategory(IntEnum):
    """Categories of multiplayer game modes."""

    UNKNOWN = -1
    """Unknown game mode category."""
    NONE = 0
    """No game mode category."""
    CAMPAIGN = 1
    """Campaign game modes."""
    FORGE = 2
    """Forge game modes."""
    ACADEMY = 3
    """Academy game modes."""
    ACADEMY_TUTORIAL = 4
    """Academy tutorial game modes."""
    ACADEMY_PRACTICE = 5
    """Academy practice game modes."""
    SLAYER = 6
    """Multiplayer slayer game modes."""
    ATTRITION = 7
    """Multiplayer attrition game modes."""
    ELIMINATION = 8
    """Multiplayer elimination game modes."""
    FIESTA = 9
    """Multiplayer fiesta game modes."""
    SWAT = 10
    """Multiplayer SWAT game modes."""
    STRONGHOLDS = 11
    """Multiplayer strongholds game modes."""
    BASTION = 12
    """Multiplayer bastion game modes."""
    KING_OF_THE_HILL = 13
    """Multiplayer king of the hill game modes."""
    TOTAL_CONTROL = 14
    """Multiplayer total control game modes."""
    CTF = 15
    """Multiplayer capture the flag game modes."""
    ASSAULT = 16
    """Multiplayer assault game modes."""
    EXTRACTION = 17
    """Multiplayer extraction game modes."""
    ODDBALL = 18
    """Multiplayer oddball game modes."""
    STOCKPILE = 19
    """Multiplayer stockpile game modes."""
    JUGGERNAUT = 20
    """Multiplayer juggernaut game modes."""
    REGICIDE = 21
    """Multiplayer regicide game modes."""
    INFECTION = 22
    """Multiplayer infection game modes."""
    ESCORT = 23
    """Multiplayer escort game modes."""
    GUN_GAME = 24
    """Multiplayer gun game game modes."""
    GRIFBALL = 25
    """Multiplayer grifball game modes."""
    RACE = 26
    """Multiplayer racing game modes."""
    PROTOTYPE = 27
    """Multiplayer prototype game modes."""
    TEST = 28
    """Test game modes."""
    TEST_ACADEMY = 29
    """Test academy game modes."""
    TEST_AUDIO = 30
    """Test audio game modes."""
    TEST_CAMPAIGN = 31
    """Test campaign game modes."""
    TEST_ENGINE = 32
    """Test engine game modes."""
    TEST_FORGE = 33
    """Test forge game modes."""
    TEST_GRAPHICS = 34
    """Test graphics game modes."""
    TEST_MULTIPLAYER = 35
    """Test multiplayer game modes."""
    TEST_SANDBOX = 36
    """Test sandbox game modes."""
    ACADEMY_TRAINING = 37
    """Academy training game modes."""
    ACADEMY_WEAPON_DRILL = 38
    """Academy weapon drill game modes."""
    LAND_GRAB = 39
    """Multiplayer land grab game modes."""
    MINIGAME = 41
    """Minigame game modes."""
    FIREFIGHT_BASTION = 42
    """Firefight king of the hill game modes."""

    @classmethod
    def _missing_(cls, value: int) -> "GameVariantCategory":
        """Return the default game mode category for an unknown value."""
        logger.error(f"Unknown game mode category: {value}")
        return cls.UNKNOWN


class InspectionResult(IntEnum):
    """Related to readiness of user-generated content?"""

    UNKNOWN = 0
    """Unknown inspection result."""
    TOLERABLE = 5
    """Tolerable inspection result."""
    STUDIO_CONTENT = 50
    """Studio content inspection result."""


class LifecycleMode(IntEnum):
    """General categories of game modes."""

    CUSTOM = 1
    """Custom games."""
    MATCHMADE = 3
    """Matchmade games."""
    LOCAL_AREA_NETWORK = 7
    """Games played on LAN."""


class MedalDifficulty(IntEnum):
    """Difficulty of medals obtainable in matchmaking."""

    # Values line up to indices in the medal metadata response content.
    NORMAL = 0
    """Normal difficulty medals. Easiest to obtain."""
    HEROIC = 1
    """Heroic difficulty medals. 2nd easiest to obtain."""
    LEGENDARY = 2
    """Legendary difficulty medals. 2nd hardest to obtain."""
    MYTHIC = 3
    """Mythic difficulty medals. Hardest to obtain."""


class MedalType(IntEnum):
    """Types of medals obtainable in matchmaking."""

    # Values line up to indices in the medal metadata response content.
    SPREE = 0
    """General or weapon-specific killing sprees."""
    MODE = 1
    """Game mode-specific medals."""
    MULTIKILL = 2
    """Kill multiple enemies in quick succession."""
    PROFICIENCY = 3
    """Medals related to player proficiency."""
    SKILL = 4
    """Medals awarded for skillful play."""
    STYLE = 5
    """Medals awarded for stylish play."""


class Outcome(IntEnum):
    """Match outcome options."""

    UNKNOWN = -1
    """Unknown match outcome."""
    NONE = 0
    """No match outcome."""
    TIE = 1
    """Match ended in a tie."""
    WIN = 2
    """Match was won."""
    LOSS = 3
    """Match was lost."""
    DID_NOT_FINISH = 4
    """Match was not finished."""
    DID_NOT_START = 5
    """Match was not started."""


class PlayerType(IntEnum):
    """Types of players."""

    UNKNOWN = -1
    """Unknown player type."""
    HUMAN = 1
    """Human players."""
    BOT = 2
    """AI players."""


class PlaylistBotDifficulty(IntEnum):
    """Bot difficulty options for matchmaking playlists."""

    AUTOMATIC = 0
    """Bot difficulty is automatically determined."""
    RECRUIT = 1
    """Recruit bot difficulty."""
    MARINE = 2
    """Marine bot difficulty."""
    ODST = 3
    """ODST bot difficulty."""
    SPARTAN = 4
    """Spartan bot difficulty."""


class PlaylistDeviceInput(IntEnum):
    """Device input options for matchmaking playlists."""

    UNKNOWN = 0
    """Unknown device input."""
    CONTROLLER = 1
    """Controller."""
    MOUSE_KEYBOARD = 2
    """Keyboard and mouse."""


class PlaylistEntrySelectionStrategy(IntEnum):
    """Selection strategies for playlist entries."""

    WEIGHTED = 0
    """Selection strategy that uses weighted probabilities."""
    NO_REPEAT = 1
    """Selection strategy that avoids repeating map mode pairs."""


class PlaylistExperience(IntEnum):
    """General categories of playlists."""

    UNKNOWN = -1
    """Unknown playlist experience."""
    NONE = 0
    """No playlist experience."""
    UNTRACKED = 1
    """Untracked playlist experience."""
    ARENA = 2
    """Arena playlists. Typically 4v4."""
    BIG_TEAM_BATTLE = 3
    """Big team battle playlists. Typically 12v12."""
    PVE = 4
    """Player vs. bots playlists."""
    FEATURED = 5
    """Featured playlists. Rotates frequently."""
    FIREFIGHT = 6
    """Firefight PvE playlist"""


class SkillResultCode(IntEnum):
    """Result codes for skill requests."""

    SUCCESS = 0
    """Skill request was successful."""
    NOT_FOUND = 1
    """Skill request failed as a requested resource was not found."""
    SERVICE_FAILURE = 2
    """Skill request failed due to a service failure."""
    SERVICE_UNAVAILABLE = 3
    """Skill request failed as the service is unavailable."""
    FORBIDDEN = 4
    """Skill request failed as the request was forbidden."""


class SubTier(Enum):
    """Sub-tiers of skill rankings.

    The `value` attribute of the sub-tier items is the index of the sub-tier,
    starting at 0 for the 1st sub-tier and incrementing as sub-tier increases.
    The `to_int` method returns the true sub-tier value, e.g., "ONE" returns 1.
    """

    ONE = 0
    """1st sub-tier."""
    TWO = 1
    """2nd sub-tier."""
    THREE = 2
    """3rd sub-tier."""
    FOUR = 3
    """4th sub-tier."""
    FIVE = 4
    """5th sub-tier."""
    SIX = 5
    """6th sub-tier. Highest sub-tier before advancing to the next tier."""

    @classmethod
    def from_int(cls, value: int) -> "SubTier":
        """Return the sub-tier from an integer value."""
        return cls(value - 1)

    def to_int(self) -> int:
        """Return the integer value of the sub-tier. For example, "ONE" => 1."""
        return self.value + 1


class Tier(StrEnum):
    """Tiers of skill rankings."""

    UNRANKED = ""
    """Not yet ranked or not applicable."""
    BRONZE = "Bronze"
    """Bronze tier. Lowest tier."""
    SILVER = "Silver"
    """Silver tier."""
    GOLD = "Gold"
    """Gold tier."""
    PLATINUM = "Platinum"
    """Platinum tier."""
    DIAMOND = "Diamond"
    """Diamond tier."""
    ONYX = "Onyx"
    """Onyx tier. Highest tier."""



"""Retrieve Halo Infinite authentication tokens."""

import datetime as dt
from dataclasses import dataclass
from typing import Any

from aiohttp import ClientSession


@dataclass(frozen=True)
class SpartanToken:
    """Response resulting from a request for a Spartan token.

    This provides the primary means of authenticating with the Halo Infinite
    API. The tokens are valid for 4 hours.

    Attributes:
        raw: The raw, deserialized JSON response.
    """

    raw: dict[str, Any]

    @property
    def token(self) -> str:
        """The spartan token value."""
        return self.raw["SpartanToken"]

    @property
    def expires_at(self) -> dt.datetime:
        """The UTC datetime at which the Spartan token expires."""
        return dt.datetime.fromisoformat(self.raw["ExpiresUtc"]["ISO8601Date"])


@dataclass(frozen=True)
class ClearanceToken:
    """A Halo Infinite clearance token response.

    A clearance token is required for some API endpoints.

    Attributes:
        raw: The raw, deserialized JSON response.
    """

    raw: dict[str, Any]

    @property
    def token(self) -> str:
        """The flight configuration ID."""
        return self.raw["FlightConfigurationId"]


async def request_spartan_token(
    session: ClientSession, halo_xsts_token: str
) -> SpartanToken:
    """Request a spartan token for authentication with Halo Infinite endpoints.

    Args:
        session: The aiohttp session to use.
        halo_xsts_token: An XSTS token scoped to Halo Waypoint.

    Returns:
        The spartan token response.
    """
    url = "https://settings.svc.halowaypoint.com/spartan-token"
    headers = {"Accept": "application/json"}
    js = {
        "Audience": "urn:343:s3:services",
        "MinVersion": "4",
        "Proof": [
            {
                "Token": halo_xsts_token,
                "TokenType": "Xbox_XSTSv3",
            }
        ],
    }
    response = await session.post(url, headers=headers, json=js)
    return SpartanToken(await response.json())


async def request_clearance_token(
    session: ClientSession, spartan_token: str
) -> ClearanceToken:
    """Request a clearance token required for some Halo Infinite endpoints.

    Args:
        session: The aiohttp session to use.
        spartan_token: The spartan token to use.

    Returns:
        The clearance token response.
    """
    url = (
        "https://settings.svc.halowaypoint.com/oban/flight-configurations/"
        "titles/hi/audiences/RETAIL/active"
    )
    headers = {"x-343-authorization-spartan": spartan_token}
    response = await session.get(url, headers=headers)
    return ClearanceToken(await response.json())



"""An Azure Active Directory (Azure AD) application."""

from dataclasses import dataclass


@dataclass(frozen=True)
class AzureApp:
    """An Azure Active Directory (Azure AD) application.

    Azure AD is a cloud-based identity and access management service from
    Microsoft. It allows you to sign in and access Microsoft resources from an
    application.

    Attributes:
        client_id: The client ID of the application.
        client_secret: The client secret of the application.
        redirect_uri: The redirect URI of the application. Defaults to
            "https://localhost".
    """

    client_id: str
    client_secret: str
    redirect_uri: str = "https://localhost"



"""Profile data services."""

from typing import Iterable

from spnkr.models.profile import User
from spnkr.responses import JsonResponse
from spnkr.services.base import BaseService
from spnkr.xuid import unwrap_xuid, wrap_xuid

_HOST = "https://profile.svc.halowaypoint.com"


class ProfileService(BaseService):
    """Profile data services."""

    async def _get_user(self, user: str) -> JsonResponse[User]:
        resp = await self._get(f"{_HOST}/users/{user}")
        return JsonResponse(resp, lambda data: User(**data))

    async def get_current_user(self) -> JsonResponse[User]:
        """Get the current user profile.

        Returns:
            The user.
        """
        return await self._get_user("me")

    async def get_user_by_gamertag(self, gamertag: str) -> JsonResponse[User]:
        """Get user profile for the given gamertag.

        Args:
            gamertag: The gamertag of the player.

        Returns:
            The user.
        """
        return await self._get_user(f"gt({gamertag})")

    async def get_user_by_id(self, xuid: str | int) -> JsonResponse[User]:
        """Get user profile for the given Xbox Live ID.

        Args:
            xuid: The Xbox Live ID of the player.

        Returns:
            The user.
        """
        return await self._get_user(wrap_xuid(xuid))

    async def get_users_by_id(
        self, xuids: Iterable[str | int]
    ) -> JsonResponse[list[User]]:
        """Get user profiles for the given list of Xbox Live IDs.

        Args:
            xuids: The Xbox Live IDs of the players.

        Returns:
            A list of users.

        Raises:
            TypeError: If `xuids` is a `str` instead of an iterable of XUIDs.
        """
        if isinstance(xuids, str):
            raise TypeError("`xuids` must be an iterable of XUIDs, got `str`")
        url = f"{_HOST}/users"
        params = {"xuids": [unwrap_xuid(x) for x in xuids]}
        resp = await self._get(url, params=params)
        return JsonResponse(resp, lambda data: [User(**u) for u in data])



"""Convert Xbox Live IDs (XUID) to and from strings and integers."""

import math
import re

from spnkr.errors import InvalidXuidError

_XUID_PATTERN = re.compile(r"^xuid\((\d{16})\)$")


def wrap_xuid(xuid: str | int) -> str:
    """Wrap an Xbox Live ID in the "xuid()" format.

    Args:
        xuid: Xbox Live ID. This can be an integer or a string representation of
            the ID. Examples of valid inputs include "xuid(1234567890123456)",
            "1234567890123456", and 1234567890123456.

    Returns:
        A wrapped version of the XUID.

    Raises:
        InvalidXuidError: If the XUID is invalid.
    """
    return f"xuid({unwrap_xuid(xuid)})"


def unwrap_xuid(xuid: str | int) -> int:
    """Get the integer value of an Xbox Live ID.

    Args:
        xuid: Xbox Live ID. This can be an integer or a string representation of
            the ID. Examples of valid inputs include "xuid(1234567890123456)",
            "1234567890123456", and 1234567890123456.

    Returns:
        The integer value of the Xbox Live ID.

    Raises:
        InvalidXuidError: If the XUID is invalid.
    """
    if isinstance(xuid, int):
        if not math.floor(math.log10(xuid)) + 1 == 16:
            # Must be a 16-digit number
            raise InvalidXuidError(xuid)
        return xuid

    xuid = xuid.strip().lower()
    if xuid.isdigit() and len(xuid) == 16:
        return int(xuid)

    match = _XUID_PATTERN.match(xuid)
    if match is None:
        raise InvalidXuidError(xuid)
    return int(match.group(1))


def wrap_xuid_or_gamertag(xuid_or_gamertag: str | int) -> str:
    """Return the value if it is a gamertag. Otherwise, wrap in "xuid()" format.

    Args:
        xuid_or_gamertag: Xbox Live ID or gamertag. This can be an integer, a
            string representation of the ID, or a gamertag. Examples of valid
            inputs include "xuid(1234567890123456)", "1234567890123456",
            1234567890123456, and "MyGamertag".

    Returns:
        A wrapped version of the XUID or the gamertag.

    Raises:
        InvalidXuidError: If the XUID is invalid.
    """
    try:
        return wrap_xuid(xuid_or_gamertag)
    except InvalidXuidError:
        if isinstance(xuid_or_gamertag, int) or not xuid_or_gamertag.strip():
            raise
        return xuid_or_gamertag.strip()  # Can safely assume it's a gamertag



"""Authenticate using OAuth2."""

import urllib.parse
from dataclasses import dataclass
from typing import Any

from aiohttp import ClientSession

from spnkr.auth.app import AzureApp
from spnkr.errors import OAuth2Error

DEFAULT_SCOPES = ["Xboxlive.signin", "Xboxlive.offline_access"]


@dataclass(frozen=True)
class OAuth2Token:
    """Response resulting from a request for an OAuth2 token.

    This is the root token used to authenticate an application. It is used to
    request an XToken, which is used to authenticate a user for Xbox Live.

    Attributes:
        raw: The raw, deserialized JSON response.
    """

    raw: dict[str, Any]

    @property
    def access_token(self) -> str:
        """The access token."""
        return self.raw["access_token"]

    @property
    def refresh_token(self) -> str:
        """The refresh token used to refresh the access token."""
        return self.raw["refresh_token"]


def generate_authorization_url(app: AzureApp) -> str:
    """Generate Windows Live Authorization URL.

    Args:
        app: The Azure AD application to authenticate.

    Returns:
        The authorization URL.
    """
    params = {
        "client_id": app.client_id,
        "response_type": "code",
        "approval_prompt": "auto",
        "scope": " ".join(DEFAULT_SCOPES),
        "redirect_uri": app.redirect_uri,
    }
    query = urllib.parse.urlencode(params)
    return f"https://login.live.com/oauth20_authorize.srf?{query}"


async def request_oauth_token(
    session: ClientSession, authorization_code: str, app: AzureApp
) -> OAuth2Token:
    """Request an OAuth2 token.

    Args:
        session: The aiohttp session to use.
        authorization_code: The authorization code to use.
        app: The Azure AD application to authenticate.

    Returns:
        The OAuth2 token response.
    """
    data = {
        "grant_type": "authorization_code",
        "code": authorization_code,
        "scope": " ".join(DEFAULT_SCOPES),
        "redirect_uri": app.redirect_uri,
    }
    return await _oauth2_token_request(session, data, app)


async def refresh_oauth_token(
    session: ClientSession, refresh_token: str, app: AzureApp
) -> OAuth2Token:
    """Refresh an OAuth2 token.

    Args:
        session: The aiohttp session to use.
        refresh_token: The refresh token to use.
        app: The Azure AD application to authenticate.

    Returns:
        The OAuth2 token response.
    """
    data = {
        "grant_type": "refresh_token",
        "scope": " ".join(DEFAULT_SCOPES),
        "refresh_token": refresh_token,
    }
    return await _oauth2_token_request(session, data, app)


async def _oauth2_token_request(
    session: ClientSession, data: dict[str, str], app: AzureApp
) -> OAuth2Token:
    """Execute an OAuth2 token request."""
    url = "https://login.live.com/oauth20_token.srf"
    request_data = {
        **data,
        "client_id": app.client_id,
        "client_secret": app.client_secret,
    }
    response = await session.post(url, data=request_data)
    if not response.ok:
        raise OAuth2Error(await response.json())
    return OAuth2Token(await response.json())



"""Base service class."""

from typing import TYPE_CHECKING, TypeAlias

from aiolimiter import AsyncLimiter

if TYPE_CHECKING:
    from aiohttp import ClientResponse, ClientSession
    from aiohttp_client_cache.response import CachedResponse
    from aiohttp_client_cache.session import CachedSession

Response: TypeAlias = "ClientResponse | CachedResponse"
Session: TypeAlias = "ClientSession | CachedSession"


class BaseService:
    """Base service class. Handles initialization and rate limiting requests."""

    def __init__(self, session: Session, requests_per_second: int = 5) -> None:
        """Initialize a service.

        Args:
            session: The authenticated aiohttp session to use.
            requests_per_second: The rate limit to use.
        """
        self._session = session
        # Setting the max rate to 1 disallows bursts.
        self._rate_limiter = AsyncLimiter(1, 1 / requests_per_second)

    async def _get(self, url: str, **kwargs) -> Response:
        """Make a GET request to `url` and return the response."""
        response = await self._session.get(url, **kwargs)
        if not hasattr(response, "from_cache"):
            # Only rate limit non-cached responses.
            await self._rate_limiter.acquire()
        response.raise_for_status()
        return response



"""Base classes for Pydantic models."""

from pydantic import BaseModel


def _to_camel_case(string: str) -> str:
    """Convert a snake case string to camelCase."""
    words = string.split("_")
    return words[0] + "".join(word.capitalize() for word in words[1:])


def _to_pascal_case(string: str) -> str:
    """Convert a snake case string to PascalCase."""
    return "".join(word.capitalize() for word in string.split("_"))


class CamelCaseModel(BaseModel, frozen=True):
    """A Pydantic model that uses camelCase for its field names."""

    model_config = {
        "alias_generator": _to_camel_case,
        "populate_by_name": True,
    }


class PascalCaseModel(BaseModel, frozen=True):
    """A Pydantic model that uses PascalCase for its field names."""

    model_config = {
        "alias_generator": _to_pascal_case,
        "populate_by_name": True,
    }



"""Services available for retrieving Halo Infinite data.

Access service instances via
[HaloInfiniteClient][spnkr.client.HaloInfiniteClient] properties.
"""

from spnkr.services.discovery_ugc import DiscoveryUgcService
from spnkr.services.economy import EconomyService
from spnkr.services.gamecms_hacs import GameCmsHacsService
from spnkr.services.profile import ProfileService
from spnkr.services.skill import SkillService
from spnkr.services.stats import StatsService

__all__ = [
    "DiscoveryUgcService",
    "EconomyService",
    "GameCmsHacsService",
    "ProfileService",
    "SkillService",
    "StatsService",
]



"""Logic for obtaining API credentials."""

from spnkr.auth.app import AzureApp
from spnkr.auth.core import authenticate_player, refresh_player_tokens
from spnkr.auth.halo import ClearanceToken, SpartanToken
from spnkr.auth.player import AuthenticatedPlayer

__all__ = [
    "AzureApp",
    "AuthenticatedPlayer",
    "ClearanceToken",
    "SpartanToken",
    "authenticate_player",
    "refresh_player_tokens",
]



"""Tools for Halo Infinite data analysis."""

import math
from enum import IntEnum
from typing import NamedTuple

from spnkr.models.refdata import (
    BotDifficulty,
    GameVariantCategory,
    LifecycleMode,
    MedalDifficulty,
    MedalType,
    Outcome,
    PlayerType,
    PlaylistExperience,
    SubTier,
    Tier,
)
from spnkr.xuid import unwrap_xuid, wrap_xuid

__all__ = [
    "wrap_xuid",
    "unwrap_xuid",
    "BOT_DIFFICULTY_MAP",
    "BOT_MAP",
    "GAME_CATEGORY_MAP",
    "LIFECYCLE_MAP",
    "MEDAL_DIFFICULTY_MAP",
    "MEDAL_TYPE_MAP",
    "MEDAL_NAME_MAP",
    "OUTCOME_MAP",
    "PLAYER_TYPE_MAP",
    "PLAYLIST_EXPERIENCE_MAP",
    "TEAM_MAP",
    "get_rank_from_csr",
    "Rank",
]


def _intenum_to_mapping(enum: type[IntEnum]) -> dict[int, str]:
    """Create a dict of value to name for an `IntEnum`."""
    return {e.value: e.name for e in enum}


BOT_DIFFICULTY_MAP: dict[int, str] = _intenum_to_mapping(BotDifficulty)
"""Mapping of bot difficulty values to names."""
GAME_CATEGORY_MAP: dict[int, str] = _intenum_to_mapping(GameVariantCategory)
"""Mapping of game variant category values to names."""
LIFECYCLE_MAP: dict[int, str] = _intenum_to_mapping(LifecycleMode)
"""Mapping of lifecycle mode values to names."""
MEDAL_DIFFICULTY_MAP: dict[int, str] = _intenum_to_mapping(MedalDifficulty)
"""Mapping of medal difficulty values to names."""
MEDAL_TYPE_MAP: dict[int, str] = _intenum_to_mapping(MedalType)
"""Mapping of medal type values to names."""
OUTCOME_MAP: dict[int, str] = _intenum_to_mapping(Outcome)
"""Mapping of match outcome values to names."""
PLAYER_TYPE_MAP: dict[int, str] = _intenum_to_mapping(PlayerType)
"""Mapping of player type values to names."""
PLAYLIST_EXPERIENCE_MAP: dict[int, str] = _intenum_to_mapping(PlaylistExperience)
"""Mapping of playlist experience values to names."""
BOT_MAP: dict[str, str] = {
    "bid(1.0)": "343 Connmando",
    "bid(2.0)": "343 Beard",
    "bid(3.0)": "343 Aloysius",
    "bid(4.0)": "343 Lemondade",
    "bid(5.0)": "343 Cream Corn",
    "bid(6.0)": "343 Rhinosaurus",
    "bid(7.0)": "343 Oscar",
    "bid(8.0)": "343 Marmot",
    "bid(9.0)": "343 Chilies",
    "bid(10.0)": "343 Cliffton",
    "bid(11.0)": "343 Ensrude",
    "bid(12.0)": "343 Godfather",
    "bid(13.0)": "343 Ritzy",
    "bid(14.0)": "343 Darkstar",
    "bid(15.0)": "343 Ham Sammich",
    "bid(16.0)": "343 Meowlnir",
    "bid(17.0)": "343 PardonMy",
    "bid(18.0)": "343 Mak",
    "bid(19.0)": "343 Ellis",
    "bid(20.0)": "343 Robot Hoida",
    "bid(21.0)": "343 Bergerton",
    "bid(22.0)": "343 Tedosaur",
    "bid(23.0)": "343 Cosmo",
    "bid(24.0)": "343 Shady Seal",
    "bid(25.0)": "343 Mumblebee",
    "bid(26.0)": "343 Total Ten",
    "bid(27.0)": "343 Byrontron",
    "bid(28.0)": "343 Forge Lord",
    "bid(29.0)": "343 Mickey",
    "bid(30.0)": "343 Lacuna",
    "bid(31.0)": "343 Hundy",
    "bid(32.0)": "343 Nando",
    "bid(33.0)": "343 Free Money",
    "bid(34.0)": "343 BF Scrub",
    "bid(35.0)": "343 Zero",
    "bid(36.0)": "343 Wiggle Cat",
    "bid(37.0)": "343 Brew Dog",
    "bid(38.0)": "343 The Thumb",
    "bid(39.0)": "343 Flippant",
    "bid(40.0)": "343 Donos",
    "bid(41.0)": "343 O Freruner",
    "bid(42.0)": "343 Tanuki",
    "bid(43.0)": "343 SpaceCase",
    "bid(44.0)": "343 GrappleMans",
    "bid(45.0)": "343 BERRYHILL",
    "bid(46.0)": "343 TooMilks",
    "bid(47.0)": "343 KaleDucky",
    "bid(48.0)": "343 BoboGan",
    "bid(49.0)": "343 Chaco",
    "bid(50.0)": "343 Colson",
    "bid(51.0)": "343 Ben Desk",
    "bid(52.0)": "343 The Referee",
    "bid(53.0)": "343 Doomfruit",
    "bid(54.0)": "343 Stone",
    "bid(55.0)": "343 Hollis",
    "bid(56.0)": "343 Razzle",
    "bid(57.0)": "343 Dazzle",
    "bid(58.0)": "343 Bachici",
    "bid(59.0)": "343 Sandwolf",
    "bid(60.0)": "343 Kubly",
}
"""Mapping of bot IDs to names. For example, 'bid(1.0)' -> '343 Connmando'."""
TEAM_MAP: dict[int, str] = {
    0: "Eagle",
    1: "Cobra",
    2: "Hades",
    3: "Valkyrie",
    4: "Rampart",
    5: "Cutlass",
    6: "Valor",
    7: "Hazard",
    8: "Observer",
}
"""Mapping of team IDs to names. For example, 0 -> 'Eagle'."""
MEDAL_NAME_MAP: dict[int, str] = {
    622331684: "Double Kill",
    2063152177: "Triple Kill",
    835814121: "Overkill",
    2137071619: "Killtacular",
    1430343434: "Killtrocity",
    3835606176: "Killamanjaro",
    2242633421: "Killtastrophe",
    3352648716: "Killpocalypse",
    3233051772: "Killionaire",
    2780740615: "Killing Spree",
    4261842076: "Killing Frenzy",
    418532952: "Running Riot",
    1486797009: "Rampage",
    865763896: "Perfection",
    3233952928: "Killjoy",
    710323196: "Nightmare",
    1720896992: "Boogeyman",
    2567026752: "Grim Reaper",
    2875941471: "Demon",
    1680000231: "Flawless Victory",
    1169390319: "Steaktacular",
    3488248720: "Stopped Short",
    976049027: "Flag Joust",
    3227840152: "Goal Line Stand",
    3011158621: "Necromancer",
    3120600565: "Immortal",
    2623698509: "Lone Wolf",
    4247875860: "Duelist",
    521420212: "Ace",
    4100966367: "Extermination",
    2477555653: "Spotter",
    1685043466: "Treasure Hunter",
    20397755: "Saboteur",
    1284032216: "Wingman",
    2926348688: "Wheelman",
    3783455472: "Gunner",
    3027762381: "Driver",
    2593226288: "Pilot",
    2278023431: "Tanker",
    2852571933: "Rifleman",
    1146876011: "Bomber",
    2648272972: "Grenadier",
    269174970: "Boxer",
    1210678802: "Warrior",
    1172766553: "Gunslinger",
    3347922939: "Scattergunner",
    4277328263: "Sharpshooter",
    2758320809: "Marksman",
    4086138034: "Heavy",
    555849395: "Bodyguard",
    548533137: "Back Smack",
    2253222811: "Nuclear Football",
    524758914: "Boom Block",
    3114137341: "Bulltrue",
    3905838030: "Cluster Luck",
    1229018603: "Dogfight",
    2418616582: "Harpoon",
    1880789493: "Mind the Gap",
    3085856613: "Ninja",
    87172902: "Odin's Raven",
    3876426273: "Pancake",
    1312042926: "Quigley",
    3160646854: "Remote Detonation",
    3059799290: "Return to Sender",
    656245292: "Rideshare",
    731054446: "Skyjack",
    3655682764: "Stick",
    1841872491: "Tag & Bag",
    1734214473: "Whiplash",
    3546244406: "Kong",
    1623236079: "Autopilot Engaged",
    670606868: "Sneak King",
    2827657131: "Windshield Wiper",
    2123530881: "Reversal",
    3934547153: "Hail Mary",
    265478668: "Nade Shot",
    4229934157: "Snipe",
    1512363953: "Perfect",
    2414983178: "Bank Shot",
    988255960: "Fire & Forget",
    4215552487: "Ballista",
    4132863117: "Pull",
    2602963073: "No Scope",
    3217141618: "Achilles Spine",
    1646928910: "Grand Slam",
    3334154676: "Guardian Angel",
    651256911: "Interlinked",
    677323068: "Death Race",
    1969067783: "Chain Reaction",
    1427176344: "360",
    641726424: "Combat Evolved",
    2396845048: "Deadly Catch",
    197913196: "Driveby",
    1211820913: "Fastball",
    3739610597: "Flyin' High",
    2625820422: "From the Grave",
    3588869844: "From the Void",
    690125105: "Grapple-jack",
    175594566: "Hold This",
    3091261182: "Last Shot",
    3475540930: "Lawnmower",
    1065136443: "Mount Up",
    1283796619: "Off the Rack",
    2861418269: "Quick Draw",
    3583966655: "Party's Over",
    2019283350: "Pineapple Express",
    1298835518: "Ramming Speed",
    1445036152: "Reclaimer",
    1169571763: "Shot Caller",
    1176569867: "Yard Sale",
    275666139: "Special Delivery",
    2967011722: "Street Sweeper",
    3732790338: "Fumble",
    781229683: "Straight Balling",
    1472686630: "Always Rotating",
    3630529364: "Clock Stop",
    221693153: "Splatter",
    3528500956: "All That Juice",
    1376646881: "Great Journey",
    629165579: "Power Outage",
    2750622016: "Breacher",
    1331361851: "Mounted & Loaded",
    2717755703: "Sole Survivor",
    1025827095: "Culling",
    88914608: "Blight",
    557309779: "Zombie Slayer",
    3467301935: "Purge",
    1155542859: "Disease",
    1447057920: "Undead Hunter",
    1064731598: "Untainted",
    1765213446: "Cleansing",
    3786134933: "Plague",
    217730222: "Hell's Janitor",
    17866865: "The Sickness",
    496411737: "Purification",
    1719203329: "Pestilence",
    2164872967: "Divine Intervention",
    3520382976: "Scourge",
    3653884673: "Apocalypse",
    394349536: "Clear Reception",
    4285712605: "Hang Up",
    2964157454: "Call Blocked",
    2426456555: "Secure Line",
    3931425309: "Signal Block",
    4007438389: "Blind Fire",
    1477806194: "Counter-snipe",
    1090931685: "Monopoly",
    580478179: "Hill Guardian",
    2848470465: "Death Cabbie",
    3169118333: "Driving Spree",
    1739996188: "Immortal Chauffeur",
    1053114074: "Clash of Kings",
    2125906504: "Big Deal",
    590706932: "Contract Killer",
    1210968206: "Watch the Throne",
}
"""Mapping of medal name IDs to medal names. For example, 622331684 -> 'Double Kill'."""


class Rank(NamedTuple):
    """A CSR tier and sub-tier.

    Attributes:
        tier: CSR tier.
        sub_tier: CSR sub-tier.
    """

    tier: Tier
    sub_tier: SubTier

    def __str__(self) -> str:
        if self.tier is Tier.ONYX:
            return self.tier.value
        return f"{self.tier.value} {self.sub_tier.to_int()}"


def get_rank_from_csr(csr: int | float) -> Rank:
    """Get the tier and sub-tier from a CSR value.

    Args:
        csr: CSR value to convert.

    Returns:
        A ranking named tuple containing `tier` and `subtier`.
    """
    csr = math.floor(csr)
    quotient, remainder = divmod(csr, 300)
    if quotient == 0:
        tier = Tier.BRONZE
    elif quotient == 1:
        tier = Tier.SILVER
    elif quotient == 2:
        tier = Tier.GOLD
    elif quotient == 3:
        tier = Tier.PLATINUM
    elif quotient == 4:
        tier = Tier.DIAMOND
    else:
        return Rank(Tier.ONYX, SubTier.ONE)
    sub_tier = SubTier.from_int((remainder // 50) + 1)
    return Rank(tier, sub_tier)
